/*---------------------------------------------------------------------------
 
  FILENAME:
        msgRouter.c
 
  PURPOSE:
        Provide a standalone message routing process to support the 
        "route by message ID" paradigm and the radmsgRouter.h API.
 
  REVISION HISTORY:
        Date            Engineer        Revision        Remarks
        12/02/2005      M.S. Teel       0               Original
 
  NOTES:
        See radmsgRouter.h for API details and usage.

  LICENSE:
        Copyright 2001-2005 Mark S. Teel. All rights reserved.
 
        Redistribution and use in source and binary forms, with or without 
        modification, are permitted provided that the following conditions 
        are met:
 
        1. Redistributions of source code must retain the above copyright 
           notice, this list of conditions and the following disclaimer.
        2. Redistributions in binary form must reproduce the above copyright 
           notice, this list of conditions and the following disclaimer in the 
           documentation and/or other materials provided with the distribution.
 
        THIS SOFTWARE IS PROVIDED BY Mark Teel ``AS IS'' AND ANY EXPRESS OR 
        IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
        WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
        DISCLAIMED. IN NO EVENT SHALL MARK TEEL OR CONTRIBUTORS BE LIABLE FOR 
        ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
        DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
        OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
        HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
        STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
        IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
        POSSIBILITY OF SUCH DAMAGE.
  
----------------------------------------------------------------------------*/

//  System include files
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>

//  Library include files
#include <radmsgRouter.h>

//  Local include files

//  global memory declarations

//  global memory referenced

//  static (local) memory declarations and methods

static MSGRTR_WORK          msgrtrWork;         // for the msg router process only

//  system initialization
static int msgrtrSysInit (MSGRTR_WORK *work, char *workingDir)
{
    struct stat     fileData;

    sprintf (work->pidFile, "%s/%s", workingDir, MSGRTR_LOCK_FILE_NAME);
    sprintf (work->fifoFile, "%s/%s", workingDir, MSGRTR_QUEUE_NAME);

    // check for our pid file, don't run if it IS there
    if (stat (work->pidFile, &fileData) == 0)
    {
        radMsgLogInit (PROC_NAME_MSGRTR, TRUE, TRUE);
        radMsgLog (PRI_CATASTROPHIC, 
                   "lock file %s exists, older copy may be running - aborting!",
                   work->pidFile);
        radMsgLogExit ();
        return ERROR;
    }

    return OK;
}

// system exit
static int msgrtrSysExit (MSGRTR_WORK *work)
{
    struct stat     fileData;

    /*  ... delete our pid file
    */
    if (stat (work->pidFile, &fileData) == 0)
    {
        unlink (work->pidFile);
    }

    return OK;
}

static void defaultSigHandler (int signum)
{
    switch (signum)
    {
        case SIGPIPE:
            signal (signum, defaultSigHandler);
            break;

        case SIGCHLD:
            wait (NULL);
            signal (signum, defaultSigHandler);
            break;

        case SIGILL:
        case SIGBUS:
        case SIGFPE:
        case SIGSEGV:
        case SIGXFSZ:
        case SIGSYS:
            // unrecoverable signal - we must exit right now!
            radMsgLog (PRI_CATASTROPHIC, "%s: recv sig %d: bailing out!", 
                       PROC_NAME_MSGRTR, signum);
            msgrtrSysExit (&msgrtrWork);
            abort ();

        default:
            // can we allow the process to exit normally?
            if (radProcessGetExitFlag())
            {
                // NO! - we gotta bail here!
                radMsgLog (PRI_HIGH, "%s: recv sig %d: exiting now!", 
                           PROC_NAME_MSGRTR, signum);

                msgrtrSysExit (&msgrtrWork);
                exit (0);                
            }

            // we can allow the process to exit normally...
            radMsgLog (PRI_HIGH, "%s: recv sig %d: exiting!", 
                       PROC_NAME_MSGRTR, signum);

            radProcessSetExitFlag ();

            signal (signum, defaultSigHandler);
            break;
    }

    return;
}

static MSGRTR_PIB *getPIB (int findpid)
{
    MSGRTR_PIB      *node;

    for (node = (MSGRTR_PIB *)radListGetFirst (&msgrtrWork.pibList);
         node != NULL;
         node = (MSGRTR_PIB *)radListGetNext (&msgrtrWork.pibList, (NODE *)node))
    {
        if (node->pid == findpid)
        {
            // got 'em!
            return node;
        }
    }

    return NULL;
}

static MSGRTR_MIIB *getMIIB (int findMsgID)
{
    MSGRTR_MIIB     *node;

    for (node = (MSGRTR_MIIB *)radListGetFirst (&msgrtrWork.miibList);
         node != NULL;
         node = (MSGRTR_MIIB *)radListGetNext (&msgrtrWork.miibList, (NODE *)node))
    {
        if (node->msgID == findMsgID)
        {
            // got 'em!
            return node;
        }
    }

    return NULL;
}

static int sendACK (MSGRTR_PIB *dest)
{
    MSGRTR_HDR          *hdr;
    MSGRTR_INTERNAL_MSG *msg;

    hdr = (MSGRTR_HDR *)radBufferGet (sizeof (*hdr) + sizeof(*msg));
    if (hdr == NULL)
    {
        radMsgLog (PRI_HIGH, "sendACK: radBufferGet failed!");
        return ERROR;
    }

    hdr->magicNumber        = MSGRTR_MAGIC_NUMBER;
    hdr->srcpid             = getpid ();
    hdr->msgID              = MSGRTR_INTERNAL_MSGID;
    hdr->length             = sizeof (*msg);

    msg = (MSGRTR_INTERNAL_MSG *)hdr->msg;
    msg->subMsgID           = MSGRTR_SUBTYPE_ACK;

    if (radProcessQueueSend (dest->queueName,
                             MSGRTR_INTERNAL_MSGID,
                             hdr,
                             sizeof (*hdr) + sizeof(*msg))
        != OK)
    {
        radMsgLog (PRI_HIGH, "sendACK: %s: radProcessQueueSend failed!",
                   dest->name);
        radBufferRls (hdr);
        return ERROR;
    }

    return OK;
}

static int sendToConsumer (MSGRTR_PIB *consumer, ULONG msgID, void *data, int length)
{
    UCHAR           *sendBfr;

    sendBfr = (UCHAR *)radBufferGet (length);
    if (sendBfr == NULL)
    {
        radMsgLog (PRI_HIGH, "sendToConsumer: radBufferGet failed!");
        return ERROR;
    }

    memcpy (sendBfr, data, length);

    if (radProcessQueueSend (consumer->queueName, msgID, sendBfr, length) != OK)
    {
        radMsgLog (PRI_HIGH, "sendToConsumer: %s: radProcessQueueSend failed!",
                   consumer->name);
        radBufferRls (sendBfr);
        consumer->rxErrors ++;
        return ERROR;
    }

    consumer->receives ++;
    msgrtrWork.transmits ++;
    return OK;
}

static void removeConsumer (MSGRTR_PIB *consumer)
{
    MSGRTR_MIIB     *node, *tempnode;
    MSGRTR_CIB      *cib;

    for (node = (MSGRTR_MIIB *)radListGetFirst (&msgrtrWork.miibList);
         node != NULL;
         node = (MSGRTR_MIIB *)radListGetNext (&msgrtrWork.miibList, (NODE *)node))
    {
        // loop through the consumer list for this msgID
        for (cib = (MSGRTR_CIB *)radListGetFirst (&node->consumers);
             cib != NULL;
             cib = (MSGRTR_CIB *)radListGetNext (&node->consumers, (NODE *)cib))
        {
            if (cib->pib == consumer)
            {
                // got 'em!
                radListRemove (&node->consumers, (NODE *)cib);
                free (cib);
                break;
            }
        }

        // if the consumer list is empty, remove the MIIB
        if (radListGetNumberOfNodes(&node->consumers) == 0)
        {
            tempnode = (MSGRTR_MIIB *)radListGetPrevious (&msgrtrWork.miibList, (NODE *)node);
            radListRemove (&msgrtrWork.miibList, (NODE *)node);
            free (node);
            node = tempnode;
        }
    }

    return;
}

static void msgHandler
(
    char                *srcQueueName,
    UINT                msgType,
    void                *msg,
    UINT                length,
    void                *userData
)
{
    MSGRTR_HDR          *hdr = (MSGRTR_HDR *)msg;
    MSGRTR_INTERNAL_MSG *intMsg = (MSGRTR_INTERNAL_MSG *)hdr->msg;
    MSGRTR_PIB          *pib;
    MSGRTR_MIIB         *miib;
    MSGRTR_CIB          *cib;

    if (hdr->magicNumber != MSGRTR_MAGIC_NUMBER)
    {
        radMsgLog (PRI_HIGH, "msgHandler: RX bad magic number 0x%8.8X", 
                   hdr->magicNumber);
        return;
    }

    if (hdr->msgID == MSGRTR_INTERNAL_MSGID)
    {
        switch (intMsg->subMsgID)
        {
            case MSGRTR_SUBTYPE_REGISTER:
            {
                if ((pib = getPIB (hdr->srcpid)) != NULL)
                {
                    // he already exists, just ACK his funky self
                    sendACK (pib);
                    return;
                }

                // let's insert this guy
                pib = (MSGRTR_PIB *)malloc (sizeof(*pib));
                if (pib == NULL)
                {
                    radMsgLog (PRI_HIGH, "msgHandler: %s: malloc PIB failed!",
                               intMsg->name);
                    return;
                }

                memset (pib, 0, sizeof(*pib));
                pib->pid = hdr->srcpid;
                strncpy (pib->name, intMsg->name, PROCESS_MAX_NAME_LEN);
                strncpy (pib->queueName, srcQueueName, QUEUE_NAME_LENGTH);

                //  attach queue
                if (radProcessQueueAttach (pib->queueName, QUEUE_GROUP_ALL) == ERROR)
                {
                    radMsgLog (PRI_HIGH, "radProcessQueueAttach %s failed!",
                               pib->queueName);
                    free (pib);
                    return;
                }

                radListAddToEnd (&msgrtrWork.pibList, (NODE *)pib);

                // finally, ACK 'em
                sendACK (pib);
                return;
            }

            case MSGRTR_SUBTYPE_DEREGISTER:
            {
                if ((pib = getPIB (hdr->srcpid)) == NULL)
                {
                    // he does not exist
                    return;
                }

                // first, remove him from the consumer lists
                removeConsumer (pib);

                // remove him from the PIB list
                radListRemove (&msgrtrWork.pibList, (NODE *)pib);

                // ACK 'em
                sendACK (pib);

                radProcessQueueDettach (pib->queueName, QUEUE_GROUP_ALL);
                free (pib);

                return;
            }

            case MSGRTR_SUBTYPE_ENABLE_MSGID:
            {
                if ((pib = getPIB (hdr->srcpid)) == NULL)
                {
                    // he does not exist
                    return;
                }

                // first, get the MIID
                miib = getMIIB (intMsg->targetMsgID);
                if (miib == NULL)
                {
                    // new msgID, create the MIIB
                    miib = (MSGRTR_MIIB *)malloc (sizeof(*miib));
                    if (miib == NULL)
                    {
                        radMsgLog (PRI_HIGH, "msgHandler: %d: malloc MIIB failed!",
                                   intMsg->targetMsgID);
                        return;
                    }

                    miib->msgID = intMsg->targetMsgID;
                    radListReset (&miib->consumers);
                    radListAddToEnd (&msgrtrWork.miibList, (NODE *)miib);
                }

                // now that we have the MIIB, add this consumer
                cib = (MSGRTR_CIB *)malloc (sizeof(*cib));
                if (cib == NULL)
                {
                    radMsgLog (PRI_HIGH, "msgHandler: %d: malloc CIB failed!",
                               intMsg->targetMsgID);
                    return;
                }

                cib->pib = pib;
                radListAddToEnd (&miib->consumers, (NODE *)cib);
                return;
            }

            case MSGRTR_SUBTYPE_DISABLE_MSGID:
            {
                if ((pib = getPIB (hdr->srcpid)) == NULL)
                {
                    // he does not exist
                    return;
                }

                // first, get the MIID
                miib = getMIIB (intMsg->targetMsgID);
                if (miib == NULL)
                {
                    // msgID does not exist...
                    return;
                }

                // now that we have the MIIB, remove this consumer
                for (cib = (MSGRTR_CIB *)radListGetFirst (&miib->consumers);
                     cib != NULL;
                     cib = (MSGRTR_CIB *)radListGetNext (&miib->consumers, (NODE *)cib))
                {
                    if (cib->pib == pib)
                    {
                        // remove him
                        radListRemove (&miib->consumers, (NODE *)cib);
                        free (cib);
                        return;
                    }
                }

                return;
            }

            case MSGRTR_SUBTYPE_DUMP_STATS:
            {
                radMsgLog (PRI_MEDIUM, "---------- Message Router Totals:  TX:%10.10u  RX:%10.10u ----------",
                           msgrtrWork.transmits, msgrtrWork.receives);
                radMsgLog (PRI_MEDIUM, "     Name     \t MSGS TX  \t MSGS RX  \t  TXERRS  \t  RXERRS");
                radMsgLog (PRI_MEDIUM, "--------------\t----------\t----------\t----------\t----------");

                // loop through the PIBs
                for (pib = (MSGRTR_PIB *)radListGetFirst (&msgrtrWork.pibList);
                     pib != NULL;
                     pib = (MSGRTR_PIB *)radListGetNext (&msgrtrWork.pibList, (NODE *)pib))
                {
                    radMsgLog (PRI_MEDIUM, "%-14s\t%10u\t%10u\t%10u\t%10u",
                               pib->name,
                               pib->transmits, 
                               pib->receives,
                               pib->txErrors,
                               pib->rxErrors);
                }

                radMsgLog (PRI_MEDIUM, "--------------------------------------------------------------------------");
                return;
            }
        }

        return;
    }

    
    //// normal message, route it ////

    // get the PIB
    if ((pib = getPIB (hdr->srcpid)) == NULL)
    {
        // he does not exist
        return;
    }

    msgrtrWork.receives ++;

    // get the proper MIIB
    miib = getMIIB (hdr->msgID);
    if (miib == NULL)
    {
        // no one to send it to
        pib->txErrors ++;
        return;
    }

    pib->transmits ++;

    // loop through the consumer list, sending it to each one of them
    for (cib = (MSGRTR_CIB *)radListGetFirst (&miib->consumers);
         cib != NULL;
         cib = (MSGRTR_CIB *)radListGetNext (&miib->consumers, (NODE *)cib))
    {
        // send it to him
        if (sendToConsumer (cib->pib, hdr->msgID, hdr->msg, hdr->length) == ERROR)
        {
            radMsgLog (PRI_HIGH, "msgHandler: %s: sendToConsumer failed!",
                       cib->pib->name);
        }
    }
            
    return;
}

static void evtHandler
(
    UINT        eventsRx,
    UINT        rxData,
    void        *userData
)
{
    return;
}

static void USAGE (void)
{
    printf ("radmrouted: invalid arguments:\n");
    printf ("USAGE: [prefix]/radmrouted [radSystemID] [workingDirectory]\n");
    
    printf ("    radSystemID         1-255, same system ID used by other processes in this group\n");
    printf ("    workingDirectory    where to store FIFO and pid files for radmrouted\n");
}

// the entry point for the message router process
int main (int argc, char *argv[])
{
    void            (*alarmHandler)(int);
    int             retVal;
    FILE            *pidfile;
    int             radSysID;

    if (argc < 3)
    {
        USAGE ();
        exit (1);
    }

    radSysID = atoi (argv[1]);
    if (radSysID < 1 || radSysID > 255)
    {
        USAGE ();
        exit (2);
    }

    memset (&msgrtrWork, 0, sizeof (msgrtrWork));
    radListReset (&msgrtrWork.pibList);
    radListReset (&msgrtrWork.miibList);

    // initialize some system stuff first
    retVal = msgrtrSysInit (&msgrtrWork, argv[2]);
    if (retVal == ERROR)
    {
        radMsgLogInit (PROC_NAME_MSGRTR, FALSE, TRUE);
        radMsgLog (PRI_CATASTROPHIC, "init failed!");
        radMsgLogExit ();
        exit (1);
    }

    msgrtrWork.radSystemID = (UCHAR)radSysID;

    /*  ... call the global radlib system init function
    */
    if (radSystemInit (msgrtrWork.radSystemID) == ERROR)
    {
        radMsgLogInit (PROC_NAME_MSGRTR, TRUE, TRUE);
        radMsgLog (PRI_CATASTROPHIC, "radSystemInit failed!");
        radMsgLogExit ();
        exit (1);
    }

    /*  ... call the radlib process init function
    */
    if (radProcessInit (PROC_NAME_MSGRTR,
                        msgrtrWork.fifoFile,
                        0,
                        TRUE,                      // TRUE for daemon
                        msgHandler,
                        evtHandler,
                        NULL)
        == ERROR)
    {
        printf ("\nradmrouted: radProcessInit failed: %s\n\n", PROC_NAME_MSGRTR);
        radSystemExit (msgrtrWork.radSystemID);
        exit (1);
    }

    msgrtrWork.myPid = getpid ();
    pidfile = fopen (msgrtrWork.pidFile, "w");
    if (pidfile == NULL)
    {
        radMsgLog (PRI_CATASTROPHIC, "lock file create failed!\n");
        radProcessExit ();
        radSystemExit (msgrtrWork.radSystemID);
        exit (1);
    }
    fprintf (pidfile, "%d", getpid ());
    fclose (pidfile);


    alarmHandler = radProcessSignalGetHandler (SIGALRM);
    radProcessSignalCatchAll (defaultSigHandler);
    radProcessSignalCatch (SIGALRM, alarmHandler);

    radMsgLog (PRI_MEDIUM, "started on radlib system %d, workdir %s",
               msgrtrWork.radSystemID, argv[2]);


    // enter normal processing
    radMsgLog (PRI_STATUS, "running...");

    while (TRUE)
    {
        // all the work happens in the message handler above
        if (radProcessWait(0) == ERROR)
        {
            break;
        }
    }


    radMsgLog (PRI_STATUS, "exiting normally...");

    radProcessSetExitFlag ();
    msgrtrSysExit (&msgrtrWork);
    radProcessExit ();
    radSystemExit (msgrtrWork.radSystemID);
    exit (0);
}

